#include "ATADriver.h"
#include "BufferManager.h"
#include "Utility.h"
#include "IOPort.h"
#include "Kernel.h"
#include "DMA.h"
#include "Chip8259A.h"

void ATADriver::ATAHandler(struct pt_regs *reg, struct pt_context *context)
{
	Buf* bp;
	Devtab* atab;
	short major = Utility::GetMajor(DeviceManager::ROOTDEV);

	BlockDevice& bdev = 
		Kernel::Instance().GetDeviceManager().GetBlockDevice(major);
	atab = bdev.d_tab;
	
	if( atab->d_active == 0 )
	{
		return;		/* û */
	}

	bp = atab->d_actf;		/* ȡж϶ӦI/OBuf */
	atab->d_active = 0;		/* ʾ豸Ѿ */

	/* I/Oִйд̿DMAǷ */
	if( ATADriver::IsError() || DMA::IsError() )
	{
		if(++atab->d_errcnt <= 10)
		{
			bdev.Start();
			return;
		}
		bp->b_flags |= Buf::B_ERROR;
	}
	
	atab->d_errcnt = 0;		/*  */
	atab->d_actf = bp->av_forw;		/* I/OȡɵI/OBuf */
	Kernel::Instance().GetBufferManager().IODone(bp);	/* I/Oƺ */
	bdev.Start();	/* I/OһI/O */
	/* 8259AжϿоƬֱEOI */
	IOPort::OutByte(Chip8259A::MASTER_IO_PORT_1, Chip8259A::EOI);
	IOPort::OutByte(Chip8259A::SLAVE_IO_PORT_1, Chip8259A::EOI);
	return;
}

void ATADriver::DevStart(struct Buf* bp)
{
	if(bp == NULL)
	{
		Utility::Panic("Invalid Buf in DevStart()!");
	}

	/* ȴϳʱδʾ */
	if(ATADriver::IsControllerReady() == 0)
	{
		Utility::Panic("Disk Hang Up!");
	}

	short minor = Utility::GetMinor(bp->b_dev);

	/* (PRD Table) */
	PhysicalRegionDescriptor prd;
	static PRDTable table;
	
	/* ΪBufʵַ */
	/* 
	 * ڽͼʼַp_addrû̬ռַС0xC0000000
	 * ԲܲöԴͬķһԵַȥ0xC0000000ʵַ:
	 * prd.SetBaseAddress((unsigned long)bp->b_addr - 0xC0000000); //Oops~
	 * p_addrдŵıѾַԲȡ·֮򱣳֡
	 */
	prd.SetBaseAddress((unsigned long)bp->b_addr & ~0xC0000000);
	prd.SetByteCount(bp->b_wcount);

	/* 
	 * õprdŵPRD Tableĵ0λã
	 * ұΪһ(ʹһprdɡ) 
	 */
	table.SetPhysicalRegionDescriptor(0, prd, true);

	DMA::Reset();		/* λDMA */

	/*  */
	IOPort::OutByte(ATADriver::NSECTOR_PORT, bp->b_wcount / BufferManager::BUFFER_SIZE);
	/* LBA28Ѱַģʽд̿ŵ0-7λ */
	IOPort::OutByte(ATADriver::BLKNO_PORT_1, bp->b_blkno & 0xFF);
	/* LBA28Ѱַģʽд̿ŵ8-15λ */
	IOPort::OutByte(ATADriver::BLKNO_PORT_2, (bp->b_blkno >> 8) & 0xFF);
	/* LBA28Ѱַģʽд̿ŵ16-23λ */
	IOPort::OutByte(ATADriver::BLKNO_PORT_3, (bp->b_blkno >> 16) & 0xFF);
	/* ATA̹ģʽĴԼLBA28е24-27λ */
	IOPort::OutByte(ATADriver::MODE_PORT, ATADriver::MODE_IDE | ATADriver::MODE_LBA28 | (minor << 4) | ((bp->b_blkno >> 24) & 0x0F) );


	/* Ƕ */
	if( (bp->b_flags & Buf::B_READ) == Buf::B_READ )
	{
		/* ߴ̿ĶдͣI/O */
		IOPort::OutByte(ATADriver::CMD_PORT, ATADriver::HD_DMA_READ);
		
		/* DMAĶдͣPRD TableʼַһDMA */
		DMA::Start(DMA::READ, table.GetPRDTableBaseAddress());
	}
	else	/* д */
	{
		IOPort::OutByte(ATADriver::CMD_PORT, ATADriver::HD_DMA_WRITE);
		
		DMA::Start(DMA::WRITE, table.GetPRDTableBaseAddress());
	}
	return;
}

int ATADriver::IsControllerReady()
{
	int ticks = 10000;
	
	while(--ticks)
	{
		unsigned char status = IOPort::InByte(ATADriver::STATUS_PORT);
		if( (status & (ATADriver::HD_DEVICE_BUSY | ATADriver::HD_DEVICE_READY)) == ATADriver::HD_DEVICE_READY )
		{
			return ticks;
		}
	}
	return 0;	/* ʱӦ */
}

bool ATADriver::IsError()
{
	unsigned char status = IOPort::InByte(ATADriver::STATUS_PORT);


	if( (status & ATADriver::HD_ERROR) == ATADriver::HD_ERROR )
	{
		return true;	/*  */
	}
	return false;	/* ûг */
}
