#include "DMA.h"
#include "IOPort.h"
#include "Utility.h"

int dmaConfig[64]; // todo

/*============================== class PhysicalRegionDescriptor ===============================*/
void PhysicalRegionDescriptor::SetBaseAddress(unsigned long phyBaseAddr)
{
	/* ȡַ[31 : 1]λдPRDӦֶ */
	this->m_MemoryRegionPhysicalBaseAddress = phyBaseAddr >> 1;
	/* ֶַεBit(0)㣬ʼַ˫ֽڶ */
	this->m_BaseAddressZeroBit = 0;
}

void PhysicalRegionDescriptor::SetByteCount(unsigned short bytes)
{
	/* DMAֽĵ[15 : 1]λдPRDӦֶ */
	this->m_ByteCount = bytes >> 1;
	/* DMAֽż */
	this->m_ByteCountZeroBit = 0;
}

void PhysicalRegionDescriptor::SetEndOfTable(bool EOT)
{
	/* EOTλ */
	this->m_EOT = EOT;
}


/*============================== class PRDTable ===============================*/
void PRDTable::SetPhysicalRegionDescriptor(int index, PhysicalRegionDescriptor &prd, bool EOT)
{
	this->m_Descriptors[index] = prd;
	this->m_Descriptors[index].SetEndOfTable(EOT);
}

unsigned long PRDTable::GetPRDTableBaseAddress()
{
	/* Ҫصַַ = Եַ - 0xC0000000 */
	return (unsigned long)this->m_Descriptors - 0xC0000000;
}


/*============================== class DMA ===============================*/
/* static member */
unsigned short DMA::COMMAND_PORT = 0;
unsigned short DMA::STATUS_PORT = 0;
unsigned short DMA::PRDTR_PORT = 0;

void DMA::Init()
{
	/* 
	 * öPCI豸ÿռ䣬ҹPCIϵIDE豸
	 * Ӷȡʼڷʸ豸DMAĶ˿ںš
	 */

	/* ϵͳPCI256,
	 * ÿPCI32PCI豸ÿPCI豸8ܺ(߼豸)*/
	int BUS = 256, DEV = 32, FUNC = 8;

	/* PCI豸ÿռַĴݼĴ˿ڡص⣬ÿPCI豸 */
	short cfg_addr_port = 0xCF8, cfg_data_port = 0xCFC;
	
	int i, j, k;
	int value;
	auto gConfig = dmaConfig;
	unsigned int config[64];	/* ÿܺŰ256ֽڵϢ */

	
	for(i = 0; i < BUS; i++)	/* öPCI */
	{
		for(j = 0; j < DEV; j++)	/* öÿϵ豸 */
		{
			for(k = 0; k < FUNC; k++)	/* öÿϵĹܺ(߼豸) */
			{
				bool find = 0;
				for(int index = 0; index < 64; index++)
				{
					unsigned int ans;
					
					/* Ҫȡÿռַcfg_addr_portÿPCI豸ַй̶Ĺ  */
					value = (0x80000000 | (i << 16) | (j << 11) | (k << 8) | (index << 2));
					IOPort::OutDWord(cfg_addr_port, value);
					
					/* Ӧÿռ */
					ans = IOPort::InDWord(cfg_data_port);
					config[index] = ans;
					gConfig[index] = ans;// todo
					
					if(ans == 0xFFFFFFFF) 
						continue;
						
					/* 
					 * index == 2ʾ256ֽÿռе0x08 - 0x0Bֽڣ
					 * жϵ0x0B,0x0AֽǷΪ0x01
					 * 0x0BֽΪ豸ֽͣڵ0x01ʾǰ豸Ϊ洢豸
					 * 0x0AֽΪ豸ֽͣڵ0x01ʾǰ豸ΪIDEӿڡ
					 */
					if( (ans & 0x01010000) == 0x01010000 && index == 2)
					{
						find = 1;
					}
				}
				
				if( find == 1 )
				{
					/* Ѿȡÿռȫ256ֽڣȡ0x20 - 0x23ֽ */
					if(config[8] & 0x1)		/* Bit(0)Ϊ1ʾ˿ڵַΪI/Oַռ */
					{
						unsigned short port_addr = config[8] & 0xFFFE;	/* ȥBit(0)Ϊ˿ڻַ */
						DMA::COMMAND_PORT = port_addr;
						DMA::STATUS_PORT = port_addr + 2;
						DMA::PRDTR_PORT = port_addr + 4;


						// we should enable Bus Master Function manually.
						//
						// see:
						//   https://www.alldatasheet.com/datasheet-pdf/pdf/66091/INTEL/82371SB.html
						// 
						// added by GTY
						auto value = (0x80000000 | (i << 16) | (j << 11) | (k << 8) | (0x04));
						IOPort::OutDWord(cfg_addr_port, value);
						IOPort::OutByte(cfg_data_port, 0x7);  // enable Bus Master Function
						// then, check the result.
						value = (0x80000000 | (i << 16) | (j << 11) | (k << 8) | (0x04));
						IOPort::OutDWord(cfg_addr_port, value);
						gConfig[1] = config[1] = IOPort::InDWord(cfg_data_port);
					}
					else	/* ΪڴӳI/Oַռ */
					{
						Utility::Panic("Error: Unsupported Memory Mapped I/O");
					}
					return;
				}
			}
		}
	}
	return;
}

void DMA::Reset()
{
	/* ֹͣǰDMA */
	IOPort::OutByte(DMA::COMMAND_PORT, DMA::STOP);
	/* DMA״̬ĴеErrorInterruptλ */
	IOPort::OutByte(DMA::STATUS_PORT, DMA::Interrupt | DMA::ERROR);
	return;
}

bool DMA::IsError()
{
	unsigned char value = IOPort::InByte(DMA::STATUS_PORT);
	

	if( (value & DMA::ERROR) == DMA::ERROR )
	{
		return true;	/*  */
	}
	return false;	/* ûг */
}

void DMA::Start(enum DMAType type, unsigned long baseAddress)
{
	/* PRD TableʼַдPRDTRĴ */
	IOPort::OutDWord(DMA::PRDTR_PORT, baseAddress);
	/* ݶдһDMA */
	IOPort::OutByte(DMA::COMMAND_PORT, int(type) | int(DMA::START));
	return;
}
