#include "FileSystem.h"
#include "Utility.h"
#include "New.h"
#include "Kernel.h"
#include "OpenFileManager.h"
#include "TimeInterrupt.h"
#include "Video.h"

/*==============================class SuperBlock===================================*/
/* ϵͳȫֳSuperBlock */
SuperBlock g_spb;

SuperBlock::SuperBlock()
{
	//nothing to do here
}

SuperBlock::~SuperBlock()
{
	//nothing to do here
}

/*==============================class Mount===================================*/
Mount::Mount()
{
	this->m_dev = -1;
	this->m_spb = NULL;
	this->m_inodep = NULL;
}

Mount::~Mount()
{
	this->m_dev = -1;
	this->m_inodep = NULL;
	//ͷڴSuperBlock
	if(this->m_spb != NULL)
	{
		delete this->m_spb;
		this->m_spb = NULL;
	}
}

/*==============================class FileSystem===================================*/
FileSystem::FileSystem()
{
	//nothing to do here
}

FileSystem::~FileSystem()
{
	//nothing to do here
}

void FileSystem::Initialize()
{
	this->m_BufferManager = &Kernel::Instance().GetBufferManager();
	this->updlock = 0;
}

void FileSystem::LoadSuperBlock()
{
	User& u = Kernel::Instance().GetUser();
	BufferManager& bufMgr = Kernel::Instance().GetBufferManager();

	for (int i = 0; i < 2; i++)
	{
		int* p = (int *)&g_spb + i * 128;

		Buf* pBuf = bufMgr.Bread(DeviceManager::ROOTDEV, FileSystem::SUPER_BLOCK_SECTOR_NUMBER + i);

		Utility::DWordCopy((int *)pBuf->b_addr, p, 128);

		bufMgr.Brelse(pBuf);
	}
	if (User::NOERROR != u.u_error)
	{
		Utility::Panic("Load SuperBlock Error....!\n");
	}

	this->m_Mount[0].m_dev = DeviceManager::ROOTDEV;
	this->m_Mount[0].m_spb = &g_spb;

	g_spb.s_flock = 0;
	g_spb.s_ilock = 0;
	g_spb.s_ronly = 0;
	g_spb.s_time = Time::time;
}

SuperBlock* FileSystem::GetFS(short dev)
{
	SuperBlock* sb;
	
	/* ϵͳװѰ豸Ϊdev豸ļϵͳSuperBlock */
	for(int i = 0; i < FileSystem::NMOUNT; i++)
	{
		if(this->m_Mount[i].m_spb != NULL && this->m_Mount[i].m_dev == dev)
		{
			/* ȡSuperBlockĵַ */
			sb = this->m_Mount[i].m_spb;
			if(sb->s_nfree > 100 || sb->s_ninode > 100)
			{
				sb->s_nfree = 0;
				sb->s_ninode = 0;
			}
			return sb;
		}
	}

	Utility::Panic("No File System!");
	return NULL;
}

void FileSystem::Update()
{
	int i;
	SuperBlock* sb;
	Buf* pBuf;

	/* һڽֱͬӷ */
	if(this->updlock)
	{
		return;
	}

	/* Update()Ļֹ */
	this->updlock++;

	/* ͬSuperBlock */
	for(i = 0; i < FileSystem::NMOUNT; i++)
	{
		if(this->m_Mount[i].m_spb != NULL)	/* MountװӦĳļϵͳ */
		{
			sb = this->m_Mount[i].m_spb;

			/* SuperBlockڴ渱ûб޸ģֱӹinodeͿ̿鱻ļϵͳֻļϵͳ */
			if(sb->s_fmod == 0 || sb->s_ilock != 0 || sb->s_flock != 0 || sb->s_ronly != 0)
			{
				continue;
			}

			/* SuperBlock޸ı־ */
			sb->s_fmod = 0;
			/* дSuperBlockʱ */
			sb->s_time = Time::time;

			/* 
			 * ΪҪдصȥSuperBlockһ黺棬ڻСΪ512ֽڣ
			 * SuperBlockСΪ1024ֽڣռ2Ҫ2д
			 */
			for(int j = 0; j < 2; j++)
			{
				/* һpָSuperBlockĵ0ֽڣڶpָ512ֽ */
				int* p = (int *)sb + j * 128;

				/* Ҫд뵽豸devϵSUPER_BLOCK_SECTOR_NUMBER + jȥ */
				pBuf = this->m_BufferManager->GetBlk(this->m_Mount[i].m_dev, FileSystem::SUPER_BLOCK_SECTOR_NUMBER + j);

				/* SuperBlockе0 - 511ֽд뻺 */
				Utility::DWordCopy(p, (int *)pBuf->b_addr, 128);

				/* ед */
				this->m_BufferManager->Bwrite(pBuf);
			}
		}
	}
	
	/* ͬ޸ĹڴInodeӦInode */
	g_InodeTable.UpdateInodeTable();

	/* Update() */
	this->updlock = 0;

	/* ӳдĻд */
	this->m_BufferManager->Bflush(DeviceManager::NODEV);
}

Inode* FileSystem::IAlloc(short dev)
{
	SuperBlock* sb;
	Buf* pBuf;
	Inode* pNode;
	User& u = Kernel::Instance().GetUser();
	int ino;	/* 䵽ĿInode */

	/* ȡӦ豸SuperBlockڴ渱 */
	sb = this->GetFS(dev);

	/* SuperBlockInode˯ߵȴ */
	while(sb->s_ilock)
	{
		u.u_procp->Sleep((unsigned long)&sb->s_ilock, ProcessManager::PINOD);
	}

	/* 
	 * SuperBlockֱӹĿInodeѿգ
	 * 뵽Inodeȶinodeб
	 * Ϊ³лж̲ܻᵼ½л
	 * пܷʸᵼ²һԡ
	 */
	if(sb->s_ninode <= 0)
	{
		/* Inode */
		sb->s_ilock++;

		/* InodeŴ0ʼⲻͬUnix V6Inode1ʼ */
		ino = -1;

		/* ζInodeеĴ̿飬пInodeInode */
		for(int i = 0; i < sb->s_isize; i++)
		{
			pBuf = this->m_BufferManager->Bread(dev, FileSystem::INODE_ZONE_START_SECTOR + i);

			/* ȡַ */
			int* p = (int *)pBuf->b_addr;

			/* ûÿInodei_mode != 0ʾѾռ */
			for(int j = 0; j < FileSystem::INODE_NUMBER_PER_SECTOR; j++)
			{
				ino++;

				int mode = *( p + j * sizeof(DiskInode)/sizeof(int) );

				/* InodeѱռãܼInode */
				if(mode != 0)
				{
					continue;
				}

				/* 
				 * inodei_mode==0ʱȷ
				 * inodeǿеģΪпڴinodeûд
				 * ,ҪڴinodeǷӦ
				 */
				if( g_InodeTable.IsLoaded(dev, ino) == -1 )
				{
					/* InodeûжӦڴ濽Inode */
					sb->s_inode[sb->s_ninode++] = ino;

					/* Ѿװ򲻼 */
					if(sb->s_ninode >= 100)
					{
						break;
					}
				}
			}

			/* Ѷ굱ǰ̿飬ͷӦĻ */
			this->m_BufferManager->Brelse(pBuf);

			/* Ѿװ򲻼 */
			if(sb->s_ninode >= 100)
			{
				break;
			}
		}
		/* ԿInodeΪȴ˯ߵĽ */
		sb->s_ilock = 0;
		Kernel::Instance().GetProcessManager().WakeUpAll((unsigned long)&sb->s_ilock);
		
		/* ڴûκοInodeNULL */
		if(sb->s_ninode <= 0)
		{
			Diagnose::Write("No Space On %d !\n", dev);
			u.u_error = User::ENOSPC;
			return NULL;
		}
	}

	/* 
	 * 沿Ѿ֤ϵͳûпInode
	 * Inodeбض¼Inodeıš
	 */
	while(true)
	{
		/* ջȡInode */
		ino = sb->s_inode[--sb->s_ninode];

		/* Inodeڴ */
		pNode = g_InodeTable.IGet(dev, ino);
		/* δܷ䵽ڴinode */
		if(NULL == pNode)
		{
			return NULL;
		}

		/* Inode,Inodeе */
		if(0 == pNode->i_mode)
		{
			pNode->Clean();
			/* SuperBlock޸ı־ */
			sb->s_fmod = 1;
			return pNode;
		}
		else	/* Inodeѱռ */
		{
			g_InodeTable.IPut(pNode);
			continue;	/* whileѭ */
		}
	}
	return NULL;	/* GCC likes it! */
}

void FileSystem::IFree(short dev, int number)
{
	SuperBlock* sb;

	sb = this->GetFS(dev);	/* ȡӦ豸SuperBlockڴ渱 */
	
	/* 
	 * ֱӹĿInode
	 * ͷŵInodeɢڴInodeС
	 */
	if(sb->s_ilock)
	{
		return;
	}

	/* 
	 * ֱӹĿInode100
	 * ͬͷŵInodeɢڴInodeС
	 */
	if(sb->s_ninode >= 100)
	{
		return;
	}

	sb->s_inode[sb->s_ninode++] = number;

	/* SuperBlock޸ı־ */
	sb->s_fmod = 1;
}

Buf* FileSystem::Alloc(short dev)
{
	int blkno;	/* 䵽Ŀд̿ */
	SuperBlock* sb;
	Buf* pBuf;
	User& u = Kernel::Instance().GetUser();

	/* ȡSuperBlockڴ渱 */
	sb = this->GetFS(dev);

	/* 
	 * д̿ڱ
	 * ڲд̿ͨ
	 * ̵Free()Alloc()ɵġ
	 */
	while(sb->s_flock)
	{
		/* ˯ֱøż */
		u.u_procp->Sleep((unsigned long)&sb->s_flock, ProcessManager::PINOD);
	}

	/* ջȡд̿ */
	blkno = sb->s_free[--sb->s_nfree];

	/* 
	 * ȡ̿Ϊ㣬ʾѷ価еĿд̿顣
	 * ߷䵽Ŀд̿Ų̿(BadBlock())
	 * ζŷд̿ʧܡ
	 */
	if(0 == blkno )
	{
		sb->s_nfree = 0;
		Diagnose::Write("No Space On %d !\n", dev);
		u.u_error = User::ENOSPC;
		return NULL;
	}
	if( this->BadBlock(sb, dev, blkno) )
	{
		return NULL;
	}

	/* 
	 * ջѿգ·䵽д̿м¼һд̿ı,
	 * һд̿ıŶSuperBlockĿд̿s_free[100]С
	 */
	if(sb->s_nfree <= 0)
	{
		/* 
		 * ˴ΪҪж̲пܷл
		 * ̨Ľ̿ܶSuperBlockĿ̿ʣᵼ²һԡ
		 */
		sb->s_flock++;

		/* ÿд̿ */
		pBuf = this->m_BufferManager->Bread(dev, blkno);

		/* Ӹô̿0ֽڿʼ¼ռ4(s_nfree)+400(s_free[100])ֽ */
		int* p = (int *)pBuf->b_addr;

		/* ȶ̿s_nfree */
		sb->s_nfree = *p++;

		/* ȡкλõݣд뵽SuperBlock̿s_free[100] */
		Utility::DWordCopy(p, sb->s_free, 100);

		/* ʹϣͷԱ㱻ʹ */
		this->m_BufferManager->Brelse(pBuf);

		/* Կд̿Ϊȴ˯ߵĽ */
		sb->s_flock = 0;
		Kernel::Instance().GetProcessManager().WakeUpAll((unsigned long)&sb->s_flock);
	}

	/* ͨ³ɹ䵽һд̿ */
	pBuf = this->m_BufferManager->GetBlk(dev, blkno);	/* Ϊô̿뻺 */
	this->m_BufferManager->ClrBuf(pBuf);	/* ջе */
	sb->s_fmod = 1;	/* SuperBlock޸ı־ */

	return pBuf;
}

void FileSystem::Free(short dev, int blkno)
{
	SuperBlock* sb;
	Buf* pBuf;
	User& u = Kernel::Instance().GetUser();

	sb = this->GetFS(dev);

	/* 
	 * SuperBlock޸ı־Էֹͷ
	 * ̿Free()ִйУSuperBlockڴ渱
	 * ޸Ľһ룬͸µSuperBlockȥ
	 */
	sb->s_fmod = 1;

	/* д̿˯ߵȴ */
	while(sb->s_flock)
	{
		u.u_procp->Sleep((unsigned long)&sb->s_flock, ProcessManager::PINOD);
	}

	/* ͷŴ̿ĺϷ */
	if(this->BadBlock(sb, dev, blkno))
	{
		return;
	}

	/* 
	 * ǰϵͳѾûп̿飬
	 * ͷŵϵͳе1̿
	 */
	if(sb->s_nfree <= 0)
	{
		sb->s_nfree = 1;
		sb->s_free[0] = 0;	/* ʹ0ǿ̿־ */
	}

	/* SuperBlockֱӹд̿ŵջ */
	if(sb->s_nfree >= 100)
	{
		sb->s_flock++;

		/* 
		 * ʹõǰFree()ҪͷŵĴ̿飬ǰһ100
		 * ̿
		 */
		pBuf = this->m_BufferManager->GetBlk(dev, blkno);	/* ΪǰҪͷŵĴ̿仺 */

		/* Ӹô̿0ֽڿʼ¼ռ4(s_nfree)+400(s_free[100])ֽ */
		int* p = (int *)pBuf->b_addr;
		
		/* д̿˵һΪ99飬ÿ鶼100 */
		*p++ = sb->s_nfree;
		/* SuperBlockĿ̿s_free[100]д뻺кλ */
		Utility::DWordCopy(sb->s_free, p, 100);

		sb->s_nfree = 0;
		/* ſ̿ġǰͷ̿顱д̣ʵ˿̿¼̿ŵĿ */
		this->m_BufferManager->Bwrite(pBuf);

		sb->s_flock = 0;
		Kernel::Instance().GetProcessManager().WakeUpAll((unsigned long)&sb->s_flock);
	}
	sb->s_free[sb->s_nfree++] = blkno;	/* SuperBlockм¼µǰͷ̿ */
	sb->s_fmod = 1;
}

Mount* FileSystem::GetMount(Inode *pInode)
{
	/* ϵͳװ */
	for(int i = 0; i <= FileSystem::NMOUNT; i++)
	{
		Mount* pMount = &(this->m_Mount[i]);

		/* ҵڴInodeӦMountװ */
		if(pMount->m_inodep == pInode)
		{
			return pMount;
		}
	}
	return NULL;	/* ʧ */
}

bool FileSystem::BadBlock(SuperBlock *spb, short dev, int blkno)
{
	return 0;
}
