#ifndef MACHINE_H
#define	MACHINE_H

#include "IDT.h"
#include "GDT.h"
#include "PageDirectory.h"
#include "TaskStateSegment.h"
#include "sys/types.h"

/*
 * MachineڷװԵײӲģʽݽṹĳ
 * 8254ʱоƬ8259AжϿоƬĳʼԼ
 * ģʽGDT, IDTݽṹĲ
 * 
 * MachineʹSingletonģʽʵ֣ϵͳں
 * ֻһʵ
 */
class Machine
{
	/* static const member */
public:
	/* ں˴ΡںݶΣûΡûݶΣTSSεѡ */
	static const unsigned int KERNEL_CODE_SEGMENT_SELECTOR = 0x08;
	static const unsigned int KERNEL_DATA_SEGMENT_SELECTOR = 0x10;
	static const unsigned int USER_CODE_SEGMENT_SELECTOR = (0x18 | 0x3);
	static const unsigned int USER_DATA_SEGMENT_SELECTOR = (0x20 | 0x3);		
	static const unsigned int TASK_STATE_SEGMENT_SELECTOR = 0x28;
	static const unsigned int TASK_STATE_SEGMENT_IDX = 0x5;	/* TSSGDTеλ */

	/* ҳĿ¼̬ҳû̬ҳڴеʼַ */
	static const unsigned long PAGE_DIRECTORY_BASE_ADDRESS = 0x200000;
	static const unsigned long KERNEL_PAGE_TABLE_BASE_ADDRESS = 0x201000;

	static const unsigned long VESA_PAGE_TABLE_BASE_ADDR = 32 * 1024 * 1024;  // place it over 32M

	static const unsigned long USER_PAGE_TABLE_BASE_ADDRESS = 0x202000;
	static const unsigned long USER_PAGE_TABLE_CNT = 2;
	
	/* ں˿ռС 4M 0xC0000000 - 0xC0400000 1 PageTable */
	static const unsigned int KERNEL_SPACE_SIZE = 0x400000;
	static const unsigned long KERNEL_SPACE_START_ADDRESS	= 0xC0000000;
	
public:
	static Machine& Instance();			/* ص̬instance */
	void LoadIDT();						/* ѽõIDTĻַͳȼؽIDTRĴ */
	void LoadGDT();						/* ѽõGDTĻַͳȼؽIDTRĴ */
	void LoadTaskRegister();

	void InitIDT();
	void InitGDT();

	/**
	 * VESA Support
	 *   added by 2051565 GTY
	 */
	void InitVESAMemoryMap(uintptr_t videoMemAddr, uintptr_t virtualMemAddr, size_t videoMemSize);

	void InitPageDirectory();
	void InitUserPageTable();
	void InitTaskStateSegment();
	void EnablePageProtection();
	
	/* property functions */
public:
	IDT& GetIDT();						/* ȡǰʹõIDT */
	GDT& GetGDT();						/* ȡǰʹõGDT */
	PageDirectory& GetPageDirectory();	/* ȡǰʹõҳĿ¼ */
	PageTable& GetKernelPageTable();	/* ȡϵͳںʹõҳmap 0xc0000000ϵַ */
	PageTable* GetUserPageTableArray();	/* ȡûҳţӳ0x2020000x203000ϣ
										    ӳ0x00000000 - 0x00800000ûַ̬ռ */
	TaskStateSegment& GetTaskStateSegment();
	
private:
	static Machine instance;	/* Machineʵ */
	
	IDT* m_IDT;
	GDT* m_GDT;
	PageDirectory* m_PageDirectory;	
	PageTable*	m_KernelPageTable;
	PageTable*	m_UserPageTable;
	TaskStateSegment* m_TaskStateSegment;
};

#endif
