#include "PEParser.h"
#include "Utility.h"
#include "PageManager.h"
#include "MemoryDescriptor.h"
#include "User.h"
#include "Kernel.h"
#include "Machine.h"

#include <libyrosstd/string.h>

PEParser::PEParser()
{
    this->EntryPointAddress = 0;
    this->sectionHeaders = 0;
}

/* ԭV6++PEParser */
PEParser::PEParser(unsigned long peAddress)
{
	this->peAddress = peAddress + 0xC0000000;   // peͷַ
}

unsigned int PEParser::Relocate(Inode* p_inode, bool sharedText)
{
	User& u = Kernel::Instance().GetUser();
	unsigned long srcAddress, desAddress;
	unsigned cnt = 0;
	unsigned int i0 = 0;

	/* Ժ̹ĶΣļжĶ */
	PageTable* pUserPageTable = Machine::Instance().GetUserPageTableArray();
	unsigned int textBegin = this->TextAddress >> 12 , textLength = this->TextSize >> 12;
	PageTableEntry* pointer = (PageTableEntry *)pUserPageTable;

	
	/*̹ĶΣĶв0*/

	int secIdxInit = !!sharedText;
	if (!sharedText)
	{
		// ޸ĶεĶд־Ϊںд׼
		for (i0 = textBegin; i0 < textBegin + textLength; i0++)
			pointer[i0].m_ReadWriter = 1;

		FlushPageDirectory();
	}

	const char* wantedSections[] = {
		".text", ".data", ".rdata", // todo: ".idata",
		".rodata", ".bss"
	};

	const int wantedSectionsSize = sizeof(wantedSections) / sizeof(wantedSections[0]);

    /* ҳִ0bssĳֵ0 */
	for (int i = 0 ; i < this->ntHeader.FileHeader.NumberOfSections; i++ )
	{
		auto& secHeader = this->sectionHeaders[i];
		bool wanted = false;
		for (int secIdx = secIdxInit ; secIdx < wantedSectionsSize; secIdx++) {
			if (strcmp(wantedSections[secIdx], secHeader.Name) == 0) {
				wanted = true;
				break;
			}
		}

		if (!wanted)
			continue;

		ImageSectionHeader* sectionHeader = &secHeader;
		int beginVM = sectionHeader->VirtualAddress + ntHeader.OptionalHeader.ImageBase;
		int size = ((sectionHeader->Misc.VirtualSize + PageManager::PAGE_SIZE - 1)>>12)<<12;
		int j;

//		if(sharedText == 0 || i != 0)
//		{
			for (j=0; j<size; j++)
			{
				unsigned char* b =(unsigned char*)(j + beginVM);
				*b = 0;
			}
//		}
	}


	/* ĶΣoptionalļȫֱĳֵ  */


	for (int i = 0 ; i < this->ntHeader.FileHeader.NumberOfSections; i++ )
	{
		auto& secHeader = this->sectionHeaders[i];
		bool wanted = false;
		for (int secIdx = secIdxInit ; secIdx < wantedSectionsSize; secIdx++) {
			if (strcmp(wantedSections[secIdx], secHeader.Name) == 0) {
				wanted = true;
				break;
			}
		}

		if (!wanted)
			continue;

		ImageSectionHeader* sectionHeader = &secHeader;
		srcAddress = sectionHeader->PointerToRawData;
		desAddress =
			this->ntHeader.OptionalHeader.ImageBase + sectionHeader->VirtualAddress;

	    u.u_IOParam.m_Base = (unsigned char*)desAddress;
	    u.u_IOParam.m_Offset = srcAddress;
	    u.u_IOParam.m_Count = sectionHeader->Misc.VirtualSize;

	    p_inode->ReadI();

		cnt += sectionHeader->Misc.VirtualSize;
	}

	if(!sharedText)
	{   //ĶҳĻֻ
		for (i0 = 0; i0 < textLength; i0++)
			pointer[i0].m_ReadWriter = 0;

		FlushPageDirectory();
	}

	KernelPageManager& kpm = Kernel::Instance().GetKernelPageManager();
	kpm.FreeMemory(PageManager::PAGE_SIZE * 2, (unsigned long)this->sectionHeaders - 0xC0000000 );
//	kpm.FreeMemory(section_size * ntHeader.FileHeader.NumberOfSections, (unsigned long)this->sectionHeaders - 0xC0000000 );
//	delete this->sectionHeaders;
	return 	cnt;
}


bool PEParser::HeaderLoad(Inode* p_inode)
{
    ImageDosHeader dos_header;
    User& u = Kernel::Instance().GetUser();
    KernelPageManager& kpm = Kernel::Instance().GetKernelPageManager();

    /*ȡdos header*/
    u.u_IOParam.m_Base = (unsigned char*)&dos_header;
    u.u_IOParam.m_Offset = 0;
    u.u_IOParam.m_Count = 0x40;
    p_inode->ReadI();       //ļIOΪReadIӡлģ

    /*ȡnt_Header*/
    //ntHeader = (ImageNTHeader*)(kpm.AllocMemory(ntHeader_size)+0xC0000000);
    u.u_IOParam.m_Base = (unsigned char*)(&this->ntHeader);
    u.u_IOParam.m_Offset = dos_header.e_lfanew;
    u.u_IOParam.m_Count = ntHeader_size;
    p_inode->ReadI();

    if ( ntHeader.Signature!=0x00004550 )
	{
		//kpm.FreeMemory(ntHeader_size, (unsigned long)ntHeader - 0xC0000000 );
        return false;
	}


    /* ԭV6++ں ȡSection tablesҳ֮٣̬òmalloc
     * ϣں  new  free 붯̬顣ڵnew񲻶ԡôš
     * sectionHeaders = new ImageSectionHeader;
     * */
    //sectionHeaders = (ImageSectionHeader*)(kpm.AllocMemory(section_size * ntHeader.FileHeader.NumberOfSections)+0xC0000000);
    sectionHeaders = (ImageSectionHeader*)(kpm.AllocMemory(PageManager::PAGE_SIZE * 2) + 0xC0000000);
    u.u_IOParam.m_Base = (unsigned char*)sectionHeaders;
    u.u_IOParam.m_Offset = dos_header.e_lfanew + ntHeader_size;
    u.u_IOParam.m_Count = section_size * ntHeader.FileHeader.NumberOfSections;
    p_inode->ReadI();

    /*
    	 * @comment hardcode gcc߼
    	 * section ˳Ϊ .text->.data->.rdata->.bss
    	 *
    */
	this->TextAddress =
		ntHeader.OptionalHeader.BaseOfCode + ntHeader.OptionalHeader.ImageBase;
	this->TextSize =
		ntHeader.OptionalHeader.BaseOfData - ntHeader.OptionalHeader.BaseOfCode;

	this->DataAddress =
		ntHeader.OptionalHeader.BaseOfData + ntHeader.OptionalHeader.ImageBase;
	this->DataSize = ntHeader.OptionalHeader.DataDirectory[1].VirtualAddress - ntHeader.OptionalHeader.BaseOfData;// todo: this->sectionHeaders[this->IDATA_SECTION_IDX].VirtualAddress - ntHeader.OptionalHeader.BaseOfData;

    StackSize = ntHeader.OptionalHeader.SizeOfStackCommit;
    HeapSize = ntHeader.OptionalHeader.SizeOfHeapCommit;

    EntryPointAddress = ntHeader.OptionalHeader.AddressOfEntryPoint +
                    ntHeader.OptionalHeader.ImageBase;

	return true;
}
