#include "Process.h"
#include "ProcessManager.h"
#include "Kernel.h"
#include "Utility.h"
#include "Machine.h"
#include "Video.h"


Process::Process()
{
	/* ʶp_statΪSNULLʶýʹ */
	this->p_stat = SNULL;
	/* 0#Wait()ʱprocess0#Ϊ */
	this->p_ppid = -1;
}

Process::~Process()
{
}


void Process::SetRun()
{
	ProcessManager& procMgr = Kernel::Instance().GetProcessManager();

	/* ˯ԭתΪ״̬ */
	this->p_wchan = 0;
	this->p_stat = Process::SRUN;
	if ( this->p_pri < procMgr.CurPri )
	{
		procMgr.RunRun++;
	}
	if ( 0 != procMgr.RunOut && (this->p_flag & Process::SLOAD) == 0 )
	{
		procMgr.RunOut = 0;
		procMgr.WakeUpAll((unsigned long)&procMgr.RunOut);
	}
}

void Process::SetPri()
{
	int priority;
	ProcessManager& procMgr = Kernel::Instance().GetProcessManager();

	priority = this->p_cpu / 16;
	priority += ProcessManager::PUSER + this->p_nice;

	if ( priority > 255 )
	{
		priority = 255;
	}
	if ( priority > procMgr.CurPri )
	{
		procMgr.RunRun++;
	}
	this->p_pri = priority;
}

bool Process::IsSleepOn(unsigned long chan)
{
	/* 鵱ǰ˯ԭǷΪchan */
	if( this->p_wchan == chan 
		&& (this->p_stat == Process::SWAIT || this->p_stat == Process::SSLEEP) )
	{
		return true;
	}
	return false;
}

void Process::Sleep(unsigned long chan, int pri)
{
	User& u = Kernel::Instance().GetUser();
	ProcessManager& procMgr = Kernel::Instance().GetProcessManager();

	if ( pri > 0 )
	{
		/* 
		 * ڽȨ˯֮ǰԼ֮յɺ
		 * źţִֹͣSleep()ͨaRetU()ֱתTrap1()
		 */
		if ( this->IsSig() )
		{
			/* returnȷaRetU()صSystemCall::Trap1()ִ֮retָ */
			aRetU(u.u_qsav);
			return;
		}
		/* 
		* ˴жϽٽ֤˯ԭchan
		* Ľ״̬ΪSSLEEP֮䲻ᷢл
		*/
		X86Assembly::CLI();
		this->p_wchan = chan;
		/* ˯ȼpriȷ̽ߡȨ˯ */
		this->p_stat = Process::SWAIT;
		this->p_pri = pri;
		X86Assembly::STI();

		if ( procMgr.RunIn != 0 )
		{
			procMgr.RunIn = 0;
			procMgr.WakeUpAll((unsigned long)&procMgr.RunIn);
		}
		/* ǰ̷CPUл̨ */
		//Diagnose::Write("Process %d Start Sleep!\n", this->p_pid);
		Kernel::Instance().GetProcessManager().Swtch();
		//Diagnose::Write("Process %d End Sleep!\n", this->p_pid);
		/* ֮ٴμź */
		if ( this->IsSig() )
		{
			/* returnȷaRetU()صSystemCall::Trap1()ִ֮retָ */
			aRetU(u.u_qsav);
			return;
		}
	}
	else
	{
		X86Assembly::CLI();
		this->p_wchan = chan;
		/* ˯ȼpriȷ̽ߡȨ˯ */
		this->p_stat = Process::SSLEEP;
		this->p_pri = pri;
		X86Assembly::STI();

		/* ǰ̷CPUл̨ */
		//Diagnose::Write("Process %d Start Sleep!\n", this->p_pid);
		Kernel::Instance().GetProcessManager().Swtch();
		//Diagnose::Write("Process %d End Sleep!\n", this->p_pid);
	}
}

void Process::Expand(unsigned int newSize)
{
	UserPageManager& userPgMgr = Kernel::Instance().GetUserPageManager();
	ProcessManager& procMgr = Kernel::Instance().GetProcessManager();
	User& u = Kernel::Instance().GetUser();
	Process* pProcess = u.u_procp;

	unsigned int oldSize = pProcess->p_size;
	p_size = newSize;
	unsigned long oldAddress = pProcess->p_addr;
	unsigned long newAddress;

	/* ͼСͷŶڴ */
	if ( oldSize >= newSize )
	{
		userPgMgr.FreeMemory(oldSize - newSize, oldAddress + newSize);
		return;
	}

	/* ͼҪѰһСnewSizeڴ */
	SaveU(u.u_rsav);
	newAddress = userPgMgr.AllocMemory(newSize);
	/* ڴʧܣʱ */
	if ( NULL == newAddress )
	{
		SaveU(u.u_ssav);
		procMgr.XSwap(pProcess, true, oldSize);
		pProcess->p_flag |= Process::SSWAP;
		procMgr.Swtch();
		/* no return */
	}
	/* ڴɹͼ񿽱ڴȻתڴ */
	pProcess->p_addr = newAddress;
	for ( unsigned int i = 0; i < oldSize; i++ )
	{
		Utility::CopySeg(oldAddress + i, newAddress + i);
	}

	/* ͷԭռõڴ */
	userPgMgr.FreeMemory(oldSize, oldAddress);
	
	X86Assembly::CLI();
	SwtchUStruct(pProcess);
	RetU();
	X86Assembly::STI();

	u.u_MemoryDescriptor.NMapToPageTable();
}

void Process::Exit()
{
	int i;
	User& u = Kernel::Instance().GetUser();
	ProcessManager& procMgr = Kernel::Instance().GetProcessManager();
	OpenFileTable& fileTable = *Kernel::Instance().GetFileManager().m_OpenFileTable;
	InodeTable& inodeTable = *Kernel::Instance().GetFileManager().m_InodeTable;

	Diagnose::Write("Process %d is exiting\n",u.u_procp->p_pid);
	/* Reset Tracing flag */
	u.u_procp->p_flag &= (~Process::STRC);

	/* ̵źŴΪ1ʾԸźκδ */
	for ( i = 0; i < User::NSIG; i++ )
	{
		u.u_signal[i] = 1;
	}

	/* رս̴ļ */
	for ( i = 0; i < OpenFiles::NOFILES; i++ )
	{
		File* pFile = NULL;
		if ( (pFile = u.u_ofiles.GetF(i)) != NULL )
		{
			fileTable.CloseF(pFile);
			u.u_ofiles.SetF(i, NULL);
		}
	}
	/*  ʲڵfderror codeu.u_errorӰִ */
	u.u_error = User::NOERROR;

	/* ݼǰĿ¼ü */
	inodeTable.IPut(u.u_cdir);

	/* ͷŸý̶ԹĶε */
	if ( u.u_procp->p_textp != NULL )
	{
		u.u_procp->p_textp->XFree();
		u.u_procp->p_textp = NULL;
	}

	/* uд뽻ȴƺ */
	SwapperManager& swapperMgr = Kernel::Instance().GetSwapperManager();
	BufferManager& bufMgr = Kernel::Instance().GetBufferManager();
	/* uĴСᳬ512ֽڣֻдppdaǰ512ֽڣuṹȫϢ */
	int blkno = swapperMgr.AllocSwap(BufferManager::BUFFER_SIZE);
	if ( NULL == blkno )
	{
		Utility::Panic("Out of Swapper Space");
	}
	Buf* pBuf = bufMgr.GetBlk(DeviceManager::ROOTDEV, blkno);
	Utility::DWordCopy((int *)&u, (int *)pBuf->b_addr, BufferManager::BUFFER_SIZE / sizeof(int));
	bufMgr.Bwrite(pBuf);

	/* ͷڴԴ */
	u.u_MemoryDescriptor.Release();
	Process* current = u.u_procp;
	UserPageManager& userPageMgr = Kernel::Instance().GetUserPageManager();
	userPageMgr.FreeMemory(current->p_size, current->p_addr);
	current->p_addr = blkno;
	current->p_stat = Process::SZOMB;

	/* Ѹ̽ƺ */
	for ( i = 0; i < ProcessManager::NPROC; i++ )
	{
		if ( procMgr.process[i].p_pid == current->p_ppid )
		{
			procMgr.WakeUpAll((unsigned long)&procMgr.process[i]);
			break;
		}
	}
	/* ûҵ */
	if ( ProcessManager::NPROC == i )
	{
		current->p_ppid = 1;
		procMgr.WakeUpAll((unsigned long)&procMgr.process[1]);
	}

	/* Լӽ̴Լĸ */
	for ( i = 0; i < ProcessManager::NPROC; i++ )
	{
		if ( current->p_pid == procMgr.process[i].p_ppid )
		{
			Diagnose::Write("My:%d 's child %d passed to 1#process",current->p_pid,procMgr.process[i].p_pid);
			procMgr.process[i].p_ppid = 1;
			if ( procMgr.process[i].p_stat == Process::SSTOP )
			{
				procMgr.process[i].SetRun();
			}
		}
	}

	procMgr.Swtch();
}

void Process::Clone(Process& proc)
{
	User& u = Kernel::Instance().GetUser();

	/* ProcessṹеĴ󲿷 */
	proc.p_size = this->p_size;
	proc.p_stat = Process::SRUN;
	proc.p_flag = Process::SLOAD;
	proc.p_uid = this->p_uid;
	proc.p_ttyp = this->p_ttyp;
	proc.p_nice = this->p_nice;
	proc.p_textp = this->p_textp;
	
	/* ӹϵ */
	proc.p_pid = ProcessManager::NextUniquePid();
	proc.p_ppid = this->p_pid;
	
	/* ʼ̵سԱ */
	proc.p_pri = 0;		/* ȷchildСȸлռCPU */
	proc.p_time = 0;
	

	/* ļƿFileṹü+1 */
	for ( int i = 0; i < OpenFiles::NOFILES; i++ )
	{
		File* pFile;
		if ( (pFile = u.u_ofiles.GetF(i)) != NULL )
		{
			pFile->f_count++;
		}
	}
	/* 
	 * GetF()u.u_ofilesеĿ룬
	 * 粻½̴(fork)ϵͳʧܡ
	 */
	u.u_error = User::NOERROR;

	/* ӶԹĶεü */
	if ( proc.p_textp != 0 )
	{
		proc.p_textp->x_count++;
		proc.p_textp->x_ccount++;
	}

	/* ӶԵǰĿ¼ü */
	u.u_cdir->i_count++;
}

//ڶջʱԶչջ
void Process::SStack()
{
	User& u = Kernel::Instance().GetUser();
	MemoryDescriptor& md = u.u_MemoryDescriptor;
	unsigned int change = 4096;
	//unsigned int change = 0;
	md.m_StackSize += change;
	unsigned int newSize = ProcessManager::USIZE + md.m_DataSize + md.m_StackSize;

	if ( false == u.u_MemoryDescriptor.EstablishUserPageTable(md.m_TextStartAddress,
						md.m_TextSize, md.m_DataStartAddress, md.m_DataSize, md.m_StackSize) )
	{
		u.u_error = User::ENOMEM;
		return;
	}

	this->Expand(newSize);
	int dst = u.u_procp->p_addr + newSize;
	unsigned int count = md.m_StackSize - change;
	while(count--)
	{
		dst--;
		Utility::CopySeg(dst - change, dst);
	}

	u.u_MemoryDescriptor.NMapToPageTable();
}


void Process::SBreak()
{
	User& u = Kernel::Instance().GetUser();
	unsigned int newEnd = u.u_arg[0];
	MemoryDescriptor& md = u.u_MemoryDescriptor;
	unsigned int newSize = newEnd - md.m_DataStartAddress;

	if (newEnd == 0)
	{
		u.u_ar0[User::EAX] = md.m_DataStartAddress + md.m_DataSize;
		return;
	}

	if ( false == u.u_MemoryDescriptor.EstablishUserPageTable(md.m_TextStartAddress, 
						md.m_TextSize, md.m_DataStartAddress, newSize, md.m_StackSize) )
	{
		//ϵͳóʱַʽءִ·ᵼ u.u_intflg == 1u.u_error޸ΪEINTR4ۺιʵϵͳʧܡ
		//aRetU(u.u_qsav);
		return;
	}

	int change = newSize - md.m_DataSize;
	md.m_DataSize = newSize;
	newSize += ProcessManager::USIZE + md.m_StackSize;

	/* ݶС */
	if ( change < 0 )
	{
		int dst = u.u_procp->p_addr + newSize - md.m_StackSize;
		int count = md.m_StackSize;
		while(count--)
		{
			Utility::CopySeg(dst - change, dst);
			dst++;
		}
		this->Expand(newSize);
	}
	/* ݶ */
	else if ( change > 0 )
	{
		this->Expand(newSize);
		int dst = u.u_procp->p_addr + newSize;
		int count = md.m_StackSize;
		while(count--)
		{
			dst--;
			Utility::CopySeg(dst - change, dst);
		}
	}
	u.u_ar0[User::EAX] = md.m_DataStartAddress + md.m_DataSize;
}

void Process::PSignal( int signal )
{
	if ( signal >= User::NSIG )
	{
		return;
	}

	/* ѾյSIGKILLźţԺź */
	if ( this->p_sig != User::SIGKILL )
	{
		this->p_sig = signal;
	}
	/* ̵PUSER(100)ΪPUSER */
	if ( this->p_pri > ProcessManager::PUSER )
	{
		this->p_pri	= ProcessManager::PUSER;
	}
	/* ̵ĴڵȨ˯ߣ份 */
	if ( this->p_stat == Process::SWAIT )
	{
		this->SetRun();
	}
}

int Process::IsSig()
{
	User& u = Kernel::Instance().GetUser();

	/* δܵź */
	if ( this->p_sig == 0 )
	{
		return 0;
	}
	/* u.u_signal[n]ΪżűʾźŽ̴ */
	else if ( (u.u_signal[this->p_sig] & 1) == 0 )
	{
		return this->p_sig;
	}
	return 0;
}

/*
extern "C" void runtime();
extern "C" void SignalHandler();
*/

void Process::PSig(struct pt_context* pContext)
{
	User& u = Kernel::Instance().GetUser();
	int signal = this->p_sig;
	/* ѽ봦̵ź */
	this->p_sig = 0;

	if ( u.u_signal[signal] != 0 )
	{
		/* յź֮ǰִϵͳڼܲErrCode */
		u.u_error = User::NOERROR;

		unsigned int old_eip = pContext->eip;

		/* ֵ̬ΪԤûSignalHandler()׵ַ */
		/*pContext->eip = ((unsigned long)SignalHandler - (unsigned long)runtime);
		pContext->esp -= 8;
		int* pInt = (int *)pContext->esp;
		*pInt = u.u_signal[signal];
		*(pInt + 1) = old_eip;*/
		pContext->eip = u.u_signal[signal];
		pContext->esp -= 4;
		int* pInt = (int *)pContext->esp;
		*pInt = old_eip;

		/* 
		 * ǰźŴӦ걾ź֮ҪΪĬ
		 * źŴΪ0ʾźŵĴʽΪֹ̡
		 */
		u.u_signal[signal] = 0;
		return;
	}

	/* u.u_signal[n]Ϊ0źŵĴʽֹ */
	u.u_procp->Exit();
}

void Process::Nice()
{
	User& u = Kernel::Instance().GetUser();
	int niceValue = u.u_arg[0];

	if (niceValue > 20)
	{
		niceValue = 20;
	}
	if (niceValue < 0 && !u.SUser())
	{
		/* ϵͳûΪС0Ľƫֵ */
		niceValue = 0;
	}
	this->p_nice = niceValue;
}

void Process::Ssig()
{
	User& u = Kernel::Instance().GetUser();

	int signalIndex = u.u_arg[0];
	unsigned long func = u.u_arg[1];

	/* ⼸źŲ */
	if ( signalIndex <= 0 || signalIndex >= User::NSIG || signalIndex == User::SIGKILL )
	{
		u.u_error = User::EINVAL;
		return;
	}
	/* úַźŴ */
	u.u_ar0[User::EAX] = u.u_signal[signalIndex];
	u.u_signal[signalIndex] = func;
	/* 嵱ǰź */
	if ( u.u_procp->p_sig == signalIndex )
	{
		u.u_procp->p_sig = 0;
	}
}





