#include "ProcessManager.h"
#include "Machine.h"
#include "User.h"
#include "Kernel.h"
#include "Video.h"
#include "Utility.h"
#include "PEParser.h"
#include "Regs.h"
#include "MemoryDescriptor.h"
#include "ELF.h"

unsigned int ProcessManager::m_NextUniquePid = 0;

ProcessManager::ProcessManager()
{
	CurPri = 0;
	RunRun = 0;
	RunIn = 0;
	RunOut = 0;
	ExeCnt = 0;
	SwtchNum = 0;
}

ProcessManager::~ProcessManager()
{
}

void ProcessManager::Initialize()
{
	//nothing to do here
}

void ProcessManager::SetupProcessZero()
{
	//ʼProcess#0ProcessUserṹ
	Process* pProcZero = &(this->process[0]);
	pProcZero->p_stat = Process::SRUN;
	pProcZero->p_flag = Process::SLOAD | Process::SSYS;
	pProcZero->p_nice = 0;
	pProcZero->p_time = 0;
	pProcZero->p_pid = NextUniquePid();
	//ppdaջ⣬ûû̬
	pProcZero->p_size = 0x1000;
	pProcZero->p_addr = PROCESS_ZERO_PPDA_ADDRESS;
	pProcZero->p_textp = NULL;

	User& u = Kernel::Instance().GetUser();
	u.u_procp = pProcZero;
	u.u_MemoryDescriptor.m_TextStartAddress = 0;
	u.u_MemoryDescriptor.m_TextSize = 0;
	u.u_MemoryDescriptor.m_DataStartAddress = 0;
	u.u_MemoryDescriptor.m_DataSize = 0;
	u.u_MemoryDescriptor.m_StackSize = 0;
	u.u_MemoryDescriptor.m_UserPageTableArray = NULL;
//	u.u_MemoryDescriptor.Initialize();
}

unsigned int ProcessManager::NextUniquePid()
{
	return ProcessManager::m_NextUniquePid++;
}

int ProcessManager::NewProc()
{
	//Diagnose::Write("Start NewProc()\n");
	Process* child = 0;
	for (int i = 0; i < ProcessManager::NPROC; i++ )
	{
		if ( process[i].p_stat == Process::SNULL )
		{
			child = &process[i];
			break;
		}
	}
	if ( !child ) 
	{
		Utility::Panic("No Proc Entry!");
	}

	User& u = Kernel::Instance().GetUser();
	Process* current = (Process*)u.u_procp;
	//Newprocֳ֣cloneprocessṹڵ
	current->Clone(*child);

	/* ҪSaveU()ֳuΪЩ̲һ
	ù */
	SaveU(u.u_rsav);

	/* ̵û̬ҳָm_UserPageTableArraypgTable */
	PageTable* pgTable = u.u_MemoryDescriptor.m_UserPageTableArray;
	u.u_MemoryDescriptor.Initialize();
	/* ̵Եַӳձӽ̣ҳĴС */
	if ( NULL != pgTable )
	{
		u.u_MemoryDescriptor.Initialize();
		Utility::MemCopy((unsigned long)pgTable, (unsigned long)u.u_MemoryDescriptor.m_UserPageTableArray, sizeof(PageTable) * MemoryDescriptor::USER_SPACE_PAGE_TABLE_CNT);
	}

	//н̵uu_procpָnew process
	//ڱƵʱֱӸu_procp
	//ַڴ治ʱ޷uӳ䵽û
	//޸u_procpĵַ
	u.u_procp = child;

	UserPageManager& userPageManager = Kernel::Instance().GetUserPageManager();

	unsigned long srcAddress = current->p_addr;
	unsigned long desAddress = userPageManager.AllocMemory(current->p_size);
	//Diagnose::Write("srcAddress %x\n", srcAddress);
	//Diagnose::Write("desAddress %x\n", desAddress);
	if ( desAddress == 0 ) /* ڴ治Ҫswap */
	{
		current->p_stat = Process::SIDL;
		/* ӽp_addrָ򸸽ͼΪӽ̻ҪԸͼΪ */
		child->p_addr = current->p_addr;
		SaveU(u.u_ssav);
		this->XSwap(child, false, 0);
		child->p_flag |= Process::SSWAP;
		current->p_stat = Process::SRUN;
	}
	else
	{
		int n = current->p_size;
		child->p_addr = desAddress;
		while (n--)
		{
			Utility::CopySeg(srcAddress++, desAddress++);
		}
	}
	u.u_procp = current;
	/* 
	 * ͼڼ䣬̵m_UserPageTableArrayָӽ̵Եַӳձ
	 * ɺָܻΪǰݵpgTable
	 */
	u.u_MemoryDescriptor.m_UserPageTableArray = pgTable;
	//Diagnose::Write("End NewProc()\n");
	return 0;
}

/* ڽлĹУûõTSS */
int ProcessManager::Swtch()
{	
	//Diagnose::Write("Start Swtch()\n");
	User& u = Kernel::Instance().GetUser();
	SaveU(u.u_rsav);

	/* 0#̨*/
	Process* procZero = &process[0];

	/* 
	 * SwtchUStruct()RetU()Ϊٽֹжϴϡ
	 * RetU()ָesp֮δָebpʱжϽᵼ
	 * espebp̵ֱָͬĺջλá good comment
	 *
	 * Ϊʲô0#̳еѡ̨Ĳ
	 * ӽлĽǶȣǫ̈̄ѡ̨ ǣʱжϡ
	 * һĩ дϵͳidleʱִӦóУԷںִйС
	 * жϣ
	 * ںidleı־  0#˯ִ̬idle()ӳ
	 *  TimeInterrupt.cppLine 82.
	 * ǣ0#ִselect()
	 *
	 */
	X86Assembly::CLI();
	SwtchUStruct(procZero);
	/* ԭĺд   RetU(u0)u0ûõ壬ɾ */
	RetU();
	X86Assembly::STI();

	/* ѡʺ̨Ľ */
	Process* selected = Select();
	//Diagnose::Write("Process id = %d Selected!\n", selected->p_pid);

	/* ̵ֳָ */
	X86Assembly::CLI();
	SwtchUStruct(selected);

	RetU();
	X86Assembly::STI();

	User& newu = Kernel::Instance().GetUser();

	newu.u_MemoryDescriptor.NMapToPageTable();
	
	/*
	 * If the new process paused because it was
	 * swapped out, set the stack level to the last call
	 * to savu(u_ssav).  This means that the return
	 * which is executed immediately after the call to aretu
	 * actually returns from the last routine which did
	 * the savu.
	 *
	 * You are not expected to understand this.
	 */
	if ( newu.u_procp->p_flag & Process::SSWAP )
	{
		newu.u_procp->p_flag &= ~Process::SSWAP;
		aRetU(newu.u_ssav);
	}
	
	/* 
	 * forkĽ̨֮ǰڱ̨ʱ1
	 * ͬʱصNewProc()ִеĵַ
	 */
	return 1;
}

void ProcessManager::Sched()
{
	Process* pSelected;
	User& u = Kernel::Instance().GetUser();
	int seconds;
	unsigned int size;
	unsigned long desAddress;

	/* 
	 * ѡڽפʱھ״̬Ľ̻
	 */
	goto loop;

sloop:
	this->RunIn++;
	u.u_procp->Sleep((unsigned long)&RunIn, ProcessManager::PSWP);

loop:
	X86Assembly::CLI();
	seconds = -1;
	for ( int i = 0; i < ProcessManager::NPROC; i++ )
	{
		if ( this->process[i].p_stat == Process::SRUN && (this->process[i].p_flag & Process::SLOAD) == 0 && this->process[i].p_time > seconds )
		{
			pSelected = &(this->process[i]);
			seconds = pSelected->p_time;
		}
	}

	/* ûзḶ̌0#˯ߵȴҪĽ */
	if ( -1 == seconds )
	{
		this->RunOut++;
		u.u_procp->Sleep((unsigned long)&RunOut, ProcessManager::PSWP);
		goto loop;
	}

	/* нҪ룬Ƿ㹻ڴ */
	X86Assembly::STI();
	/* ̻ҪڴС */
	size = pSelected->p_size;
	/* 
	 * ڹĶΣûнͼڴУøĶεḶ̌
	 * ĶβڴУʱҪĶڽеĸ
	 */
	if ( pSelected->p_textp != NULL && 0 == pSelected->p_textp->x_ccount )
	{
		size += pSelected->p_textp->x_size;
	}
	/* ڴɹʵʻ */
	desAddress = Kernel::Instance().GetUserPageManager().AllocMemory(size);
	if ( NULL != desAddress )
	{
		goto found2;
	}

	/*
	 * ڴʧ£ڴн̣ڳռ䡣
	 * ԭ򣺴׵ѣνȨ˯״̬(SWAIT)-->
	 * ͣ״̬(SSTOP)-->Ȩ˯״̬(SSLEEP)-->״̬(SRUN)̻
	 */
	X86Assembly::CLI();
	for ( int i = 0; i < ProcessManager::NPROC; i++ )
	{

		bool pFlagIsSLOAD = (this->process[i].p_flag & (int(Process::SSYS) | int(Process::SLOCK) | int(Process::SLOAD))) == int(Process::SLOAD);

		bool statIsSWAITOrSSTOP = (this->process[i].p_stat == Process::SWAIT || this->process[i].p_stat == Process::SSTOP);

		if (pFlagIsSLOAD && statIsSWAITOrSSTOP)
		{
			goto found1;
		}
	}

	/* 
	 * ڻȨ˯״̬(SSLEEP)״̬(SRUN)̶ڳڴ֮ǰ
	 * ڽפʱǷѴﵽ3룬軻
	 */
	if ( seconds < 3 )
	{
		goto sloop;
	}

	seconds = -1;
	for ( int i = 0; i < ProcessManager::NPROC; i++ )
	{

		bool pFlagIsSLOAD = (this->process[i].p_flag & (int(Process::SSYS) | int(Process::SLOCK) | int(Process::SLOAD))) == int(Process::SLOAD);
		bool pStatIsSWAITOrSSTOP = this->process[i].p_stat == Process::SWAIT || this->process[i].p_stat == Process::SSTOP;

		if ( pFlagIsSLOAD && pStatIsSWAITOrSSTOP && pSelected->p_time > seconds ) {
			pSelected = &(this->process[i]);
			seconds = pSelected->p_time;
		}
	}

	/* ҪSSLEEPSRUN״̬̣ȼýפڴʱǷ񳬹2룬軻 */
	if ( seconds < 2 )
	{
		goto sloop;
	}

	/* pSelectedָıѡн */
found1:
	X86Assembly::STI();
	pSelected->p_flag &= ~Process::SLOAD;
	this->XSwap(pSelected, true, 0);
	/* ڳڴռٴγԻ */
	goto loop;

	/* Ѿ㹻ڴ棬ʵʵĻ */
found2:
	BufferManager& bufMgr = Kernel::Instance().GetBufferManager();
	/* 
	* ڹĶΣûнͼڴУøĶεḶ̌
	* ĶβڴУʱҪĶڽеĸ
	*/
	if ( pSelected->p_textp != NULL )
	{
		Text* pText = pSelected->p_textp;
		if ( pText->x_ccount == 0 )
		{
			/* ΪĶΣͽppdaݶΡջڽǷֿŵģȻ빲Ķ */
			if ( bufMgr.Swap(pText->x_daddr, desAddress, pText->x_size, Buf::B_READ) == false )
			{
				goto err;
			}
			/* Ķڴеʼַ */
			pText->x_caddr = desAddress;
			desAddress += pText->x_size;
		}
		pText->x_ccount++;
	}
	/* ʣಿͼppdaݶΡջ */
	if ( bufMgr.Swap(pSelected->p_addr /* blkno */, desAddress, pSelected->p_size, Buf::B_READ) == false )
	{
		goto err;
	}
	Kernel::Instance().GetSwapperManager().FreeSwap(pSelected->p_size, pSelected->p_addr /* blkno */);
	pSelected->p_addr = desAddress;
	pSelected->p_flag |= Process::SLOAD;
	pSelected->p_time = 0;
	goto loop;

err:
	Utility::Panic("Swap Error");
}

void ProcessManager::Wait()
{
	int i;
	bool hasChild = false;
	User& u = Kernel::Instance().GetUser();
	SwapperManager& swapperMgr = Kernel::Instance().GetSwapperManager();
	BufferManager& bufMgr = Kernel::Instance().GetBufferManager();
	
	Diagnose::Write("Process %d finding dead son. They are ",u.u_procp->p_pid);
	while(true)
	{
		for ( i = 0; i < NPROC; i++ )
		{
			if ( u.u_procp->p_pid == process[i].p_ppid )
			{
				Diagnose::Write("Process %d (Status:%d)  ",process[i].p_pid,process[i].p_stat);
				hasChild = true;
				/* ˯ߵȴֱӽ̽ */
				if( Process::SZOMB == process[i].p_stat )
				{
					/* wait()ϵͳ÷ӽ̵pid */
					u.u_ar0[User::EAX] = process[i].p_pid;

					process[i].p_stat = Process::SNULL;
					process[i].p_pid = 0;
					process[i].p_ppid = -1;
					process[i].p_sig = 0;
					process[i].p_flag = 0;

					/* swapperӽuṹ */
					Buf* pBuf = bufMgr.Bread(DeviceManager::ROOTDEV, process[i].p_addr);
					swapperMgr.FreeSwap(BufferManager::BUFFER_SIZE, process[i].p_addr);
					User* pUser = (User *)pBuf->b_addr;

					/* ӽ̵ʱӵ */
					u.u_cstime += pUser->u_cstime +	pUser->u_stime;
					u.u_cutime += pUser->u_cutime + pUser->u_utime;

					int* pInt = (int *)u.u_arg[0];
					/* ȡӽexit(int status)ķֵ */
					*pInt = pUser->u_arg[0];

					/* ˴ûBrelse()ϵͳᷢʲô-_- */
					bufMgr.Brelse(pBuf);
					Diagnose::Write("end wait\n");
					return;
				}
			}
		}
		if (true == hasChild)
		{
			/* ˯ߵȴֱӽ̽ */
			Diagnose::Write("wait until child process Exit! ");
			u.u_procp->Sleep((unsigned long)u.u_procp, ProcessManager::PWAIT);
			Diagnose::Write("end sleep\n");
			continue;	/* صwhile(true)ѭ */
		}
		else
		{
			/* Ҫȴӽ̣ó룬wait() */
			u.u_error = User::ECHILD;
			break;	/* Get out of while loop */
		}
	}
}

void ProcessManager::Fork()
{
	User& u = Kernel::Instance().GetUser();
	Process* child = NULL;;

	/* ѰҿеprocessΪӽ̵Ľ̿ƿ */
	for ( int i = 0; i < ProcessManager::NPROC; i++ )
	{
		if ( this->process[i].p_stat == Process::SNULL )
		{
			child = &this->process[i];
			break;
		}
	}
	if ( child == NULL )
	{
		/* ûпprocess */
		u.u_error = User::EAGAIN;
		return;
	}

	if ( this->NewProc() )	/* ӽ̷1̷0 */
	{
		/* ӽfork()ϵͳ÷0 */
		u.u_ar0[User::EAX] = 0;
		u.u_cstime = 0;
		u.u_stime = 0;
		u.u_cutime = 0;
		u.u_utime = 0;
	}
	else
	{
		/* ̽fork()ϵͳ÷ӽPID */
		u.u_ar0[User::EAX] = child->p_pid;
	}

	return;
}

extern "C" void runtime();
extern "C" void ExecShell();

/* ڸҳΪ V6  execʵ֡ȱ֧ ISUID  */
void ProcessManager::Exec()
{
	Inode* pInode;
	Text* pText;
	User& u = Kernel::Instance().GetUser();
	FileManager& fileMgr = Kernel::Instance().GetFileManager();
	UserPageManager& userPgMgr = Kernel::Instance().GetUserPageManager();
	KernelPageManager& kernelPgMgr = Kernel::Instance().GetKernelPageManager();
	BufferManager& bufMgr = Kernel::Instance().GetBufferManager();

	Diagnose::Write("Process %d execing\n",u.u_procp->p_pid);
	pInode = fileMgr.NameI(FileManager::NextChar, FileManager::OPEN);
	if ( NULL == pInode )	//Ŀ¼ʧ
	{
		return;
	}

	/* ͬʱͼĻĽƣȽ˯ */
	while( this->ExeCnt >= NEXEC )
	{
		u.u_procp->Sleep((unsigned long)&ExeCnt, ProcessManager::EXPRI);
	}
	this->ExeCnt++;

	/* ̱ӵпִļִȨޣұִеֻһļ */
	if ( fileMgr.Access(pInode, Inode::IEXEC) || (pInode->i_mode & Inode::IFMT) != 0 )
	{
		fileMgr.m_InodeTable->IPut(pInode);
		if ( this->ExeCnt >= NEXEC )
		{
			WakeUpAll((unsigned long)&ExeCnt);
		}
		this->ExeCnt--;
		return;
	}

	PEParser peParser;
	ELF::Parser elfParser;
	bool isPE = peParser.HeaderLoad(pInode);
	bool isELF = false;
	if (!isPE) {
		isELF = elfParser.load(pInode) == 0;
	}

    if ( !isPE && !isELF )
    {
        fileMgr.m_InodeTable->IPut(pInode);
        return;
    }

 	/* ȡPEͷṹõĶεʼַ */
	auto textAddr = isPE ? peParser.TextAddress : elfParser.textAddr;
	u.u_MemoryDescriptor.m_TextStartAddress = textAddr;

	auto textSize = isPE ? peParser.TextSize : elfParser.textSize;
	u.u_MemoryDescriptor.m_TextSize = textSize;

	/* ݶεʼַ */
	auto dataAddr = isPE ? peParser.DataAddress : elfParser.dataAddr;
	u.u_MemoryDescriptor.m_DataStartAddress = dataAddr;

	auto dataSize = isPE ? peParser.DataSize : elfParser.dataSize;
	u.u_MemoryDescriptor.m_DataSize = dataSize;

	/* ջγʼ */
	auto stackSize = isPE ? peParser.StackSize : elfParser.stackSize;
	u.u_MemoryDescriptor.m_StackSize = stackSize;
	
	if ( textSize + dataSize + stackSize  + PageManager::PAGE_SIZE > MemoryDescriptor::USER_SPACE_SIZE - textAddr)
	{
		fileMgr.m_InodeTable->IPut(pInode);
		u.u_error = User::ENOMEM;
		return;
	}

	/* 
	 * ڴڴûҪĲargcargv[]Щexec()ϵͳô룬
	 * λڽͼĻǰûջУݵfakeStackУȻͷԭͼ
	 * ½ͼ֮ٽfakeStackеıݲ½̵ûջС
	 */
	//unsigned long fakeStack = kernelPgMgr.AllocMemory(parser.StackSize);
	int allocLength = (stackSize + PageManager::PAGE_SIZE * 2 - 1) >> 13 << 13;
	unsigned long fakeStack = kernelPgMgr.AllocMemory(allocLength);

	int argc = u.u_arg[1];
	char** argv = (char **)u.u_arg[2];

	/* espλջ */
	unsigned int esp = MemoryDescriptor::USER_SPACE_SIZE;
	/* ʹú̬ҳӳ䣬ַϼ0xC0000000Եַ */
	unsigned long desAddress = fakeStack + allocLength + 0xC0000000;
	//unsigned long desAddress = fakeStack + parser.StackSize + 0xC0000000;
	int length;

	/* argv[]ָָвַ */
	for (int i = 0; i < argc; i++ )
	{
		length = 0;
		/* ַȣlength'\0' */
		while( NULL != argv[i][length] )
		{
			length++;
		}
		desAddress = desAddress - (length + 1);
		/* ʱ'\0'һ𿽱ȥ */
		Utility::MemCopy((unsigned long)argv[i], desAddress, length + 1);
		/* ַ½ͼûջеʼλôargv[i]ûջλڽ߼ַռ0x800000ĵײ */
		esp = esp - (length + 1);
		argv[i] = (char *)esp;
	}

	/* ŵintֵ16ֽڱ߽ */
	desAddress = desAddress & 0xFFFFFFF0;
	esp = esp & 0xFFFFFFF0;

	/* argcargv[] */
	int endValue = 0;
	desAddress -= sizeof(endValue);
	esp -= sizeof(endValue);
	/* ûջдendValueΪargv[]Ľ */
	Utility::MemCopy((unsigned long)&endValue, desAddress, sizeof(endValue));

	desAddress -= argc * sizeof(int);
	esp -= argc * sizeof(int);
	/* дargv[] */
	Utility::MemCopy((unsigned long)argv, desAddress, argc * sizeof(int));

	/* endValueָǰջargv[]ʼַargv[]ջϺǰջַ */
	endValue = esp;
	desAddress -= sizeof(int);
	esp -= sizeof(int);
	Utility::MemCopy((unsigned long)&endValue, desAddress, sizeof(int));

	/* ջargc */
	desAddress -= sizeof(int);
	esp -= sizeof(int);
	Utility::MemCopy((unsigned long)&argc, desAddress, sizeof(int));	/* Done! */


	/* ͷԭͼĹĶΣݶΣջ */
	if ( u.u_procp->p_textp != NULL )
	{
		u.u_procp->p_textp->XFree();
		u.u_procp->p_textp = NULL;
	}
	u.u_procp->Expand(ProcessManager::USIZE);

	pText = NULL;
	/* һTextṹߺ̹ͬһĶ */
	for ( int i = 0; i < ProcessManager::NTEXT; i++ )
	{
		if ( NULL == this->text[i].x_iptr )     /* ҵĵһtextṹ */
		{
			if ( NULL == pText )
			{
				pText = &(this->text[i]);
			}
		}
		else if ( pInode == this->text[i].x_iptr )		/* ⲻһtextṹһtextṹָĿִļexecϵͳҪִеӦó */
		{
			this->text[i].x_count++;
			this->text[i].x_ccount++;
			u.u_procp->p_textp = &(this->text[i]);
			pText = NULL;	/* ̹ͬһĶΣpText㣬ָһTextṹ */
			break;
		}
	}


	bool sharedText = false;

	/* ûпɹֳTextṹӦʼ */
	if ( NULL != pText )
	{
		/* 
		 * ˴i_count++ƽXFree()еIPut(x_iptr)ֻExec()ʼ
		 * NameI()IGet()ԼExec()βIPut()ͷexeļInodeصInode
		 * £̺ܿҲExec()ȡInodeǡ֮ǰصexeļͷŵInode
		 * жϣpInode (ǰexeӦInode) == this->text[i].x_iptr(֮ǰexeļInode)
		 * º֮ǰ̹ͬһTextṹͬһĶΣʵϱĳ
		 */
		pInode->i_count++;

		pText->x_ccount = 1;
		pText->x_count = 1;
		pText->x_iptr = pInode;
		pText->x_size = u.u_MemoryDescriptor.m_TextSize;
		/* ΪĶηڴ棬ĶݵĶҪȵҳӳ֮ٴmapAddressַʼexeļж */
		pText->x_caddr = userPgMgr.AllocMemory(pText->x_size);
		pText->x_daddr = Kernel::Instance().GetSwapperManager().AllocSwap(pText->x_size);
		/* uTextṹĹϵ */
		u.u_procp->p_textp = pText;
	}
	else
	{
		pText = u.u_procp->p_textp;
		sharedText = true;
	}

	unsigned int newSize = ProcessManager::USIZE + u.u_MemoryDescriptor.m_DataSize + u.u_MemoryDescriptor.m_StackSize;
	/* ͼUSIZEΪUSIZE + dataSize + stackSize */
	u.u_procp->Expand(newSize);

	/* ĶΡݶΡջγȽԵַӳձصҳ */
	u.u_MemoryDescriptor.EstablishUserPageTable(textAddr, textSize, dataAddr, dataSize, stackSize);

	/* exeļζ.textΡ.dataΡ.rdataΡ.bss */
	if (isPE) {
		peParser.Relocate(pInode, sharedText);
	} else {
		elfParser.relocate(pInode, sharedText);
	}

	/* .textswap */
	if(!sharedText)
	{
		u.u_procp->p_flag |= Process::SLOCK;
		bufMgr.Swap(pText->x_daddr, pText->x_caddr, pText->x_size, Buf::B_WRITE);
		u.u_procp->p_flag &= ~Process::SLOCK;
	}

	/* fakeStackбݵûջƵ½ͼûջ */
	//Utility::MemCopy(fakeStack | 0xC0000000, MemoryDescriptor::USER_SPACE_SIZE - parser.StackSize, parser.StackSize);
	Utility::MemCopy((fakeStack + (unsigned long)(allocLength) - stackSize) | 0xC0000000, MemoryDescriptor::USER_SPACE_SIZE - stackSize, stackSize);
	/* ͷڶexeļͱûջڴ棺mapAddressfakeStack */
	kernelPgMgr.FreeMemory(allocLength, fakeStack);

	/* 
	  * runtime()SignalHandler()ûַ̬ռ0x00000000Եַruntime()
	  * ring0˳ring3Ȩִ֮еĴ룬SignalHandler()Ϊ̵źŴڣ
	  * þźŵHandlerÿһ0x00000000ԵַӦһݶruntime()SignalHandler()
	  * 
	  */
//	unsigned char* runtimeSrc = (unsigned char*)runtime;
//	unsigned char* runtimeDst = 0x00000000;
//	for (unsigned int i = 0; i < (unsigned long)ExecShell - (unsigned long)runtime; i++)
//	{
//		*runtimeDst++ = *runtimeSrc++;
//	}

	/* ͷInodeExeCntֵ */
	fileMgr.m_InodeTable->IPut(pInode);
	if ( this->ExeCnt >= NEXEC )
	{
		WakeUpAll((unsigned long)&ExeCnt);
	}
	this->ExeCnt--;

	/* Ĭϵķʽź  */
	for (int i = 0; i < u.NSIG ; i++)
	{
		u.u_signal[i] = 0;
	}

	/* 0ͨüĴ  */
	for (int i = User::EAX - 4; i < User::EAX - 4*7 ; i = i - 4)
	{
		u.u_ar0[i] = 0;     /* ±д  User::EAX + i ɶҪǿһЩٶˡС٣׷ٶȰ */
	}

	/* exeڵַջֳеEAXΪϵͳ÷ֵruntimeҪ  */
	u.u_ar0[User::EAX] = isPE ? peParser.EntryPointAddress : elfParser.entryPointAddr;
	
	/* Exec()ϵͳõ˳ʹ֮˳ring3ʱʼִuser code */
	struct pt_context* pContext = (struct pt_context *)u.u_arg[4];
	pContext->eip = 0x00000000;	/* ˳ring3Ȩ´Եַ0x00000000runtime()ʼִ */
	//pContext->eip = parser.EntryPointAddress;
	pContext->xcs = Machine::USER_CODE_SEGMENT_SELECTOR;
	pContext->eflags = 0x200;	/* Ƿ۸޹ؽҪ */
	pContext->esp = esp;
	pContext->xss = Machine::USER_DATA_SEGMENT_SELECTOR;
}

Process* ProcessManager::Select ()
{
	/* ǰһѡ̨ */
	static int lastSelect = 0;
	
	while (true)
	{
		int priority = 256;
		int best = -1;	/* ҵ̨ */

		this->RunRun = 0;

		/* ȼߵĿн */
		for ( int count = 0; count < NPROC ; count++ )
		{
			/* һαѡн̵һʼػɨ裬ÿδ0#̿ʼ֤̻ */
			int i = (lastSelect + 1 + count) % NPROC;
			if ( Process::SRUN == process[i].p_stat && (process[i].p_flag & Process::SLOAD) != 0 )
			{
				if ( process[i].p_pri < priority )
				{
					best = i;
					priority = process[i].p_pri;
				}
			}
		}
		if ( -1 == best )
		{
			__asm__ __volatile__("hlt");
			continue;
		}

		SwtchNum++;
		if ( SwtchNum & 0x80000000 ) 
		{
			SwtchNum = 0;	/* ΪΪ */
		}
		/* ѡȼߵĿн */
		this->CurPri = priority;
		lastSelect = best;
		//Diagnose::Write("Process %d is running!",best);
		return &process[best];

	}
}

void ProcessManager::Kill()
{
	User& u = Kernel::Instance().GetUser();
	int pid = u.u_arg[0];
	int signal = u.u_arg[1];
	bool flag = false;

	for ( int i = 0; i < ProcessManager::NPROC; i++ )
	{
		/* źŸ */
		if ( u.u_procp == &process[i] )
		{
			continue;
		}
		/* źŵĽշĿ̣Ѱ */
		if ( pid != 0 && process[i].p_pid != pid)
		{
			continue;
		}
		/* pidΪ0źŷ뷢ͽͬһն˵н̣0#̲ */
		if ( pid == 0 && (process[i].p_ttyp != u.u_procp->p_ttyp || i == 0 ) )
		{
			continue;
		}
		/* ǳûҪ͡սu.uidͬɸû̷ź */
		if ( u.u_uid != 0 && u.u_uid != process[i].p_uid )
		{
			continue;
		}
		flag = true;
		/* źŷ͸Ŀ */
		process[i].PSignal(signal);
	}
	if ( false == flag )
	{
		u.u_error = User::ESRCH;
	}
}

void ProcessManager::WakeUpAll(unsigned long chan)
{
	/* ϵͳchan˯ߵĽ */
	for(int i = 0; i < ProcessManager::NPROC; i++)
	{
		if( this->process[i].IsSleepOn(chan) )
		{
			this->process[i].SetRun();
		}
	}
}

void ProcessManager::XSwap( Process* pProcess, bool bFreeMemory, int size )
{
	if ( 0 == size)
	{
		size = pProcess->p_size;
	}

	/* blkno¼䵽Ľʼ */
	int blkno = Kernel::Instance().GetSwapperManager().AllocSwap(pProcess->p_size);
	if ( 0 == blkno )
	{
		Utility::Panic("Out of Swapper Space");
	}
	/* ݼͼڴУøĶεĽ */
	if ( pProcess->p_textp != NULL )
	{
		pProcess->p_textp->XccDec();
	}
	/* ֹͬһͼظ */
	pProcess->p_flag |= Process::SLOCK;
	if ( false == Kernel::Instance().GetBufferManager().Swap(blkno, pProcess->p_addr, size, Buf::B_WRITE) )
	{
		Utility::Panic("Swap I/O Error");
	}
	if ( bFreeMemory )
	{
		Kernel::Instance().GetUserPageManager().FreeMemory(size, pProcess->p_addr);
	}
	/* ѽͼڽʼż¼p_addrУSLOAD0̽ϵĽ */
	pProcess->p_addr = blkno;
	pProcess->p_flag &= ~(Process::SLOAD | Process::SLOCK);
	/* һα򻻳ڳ򽻻פʱ䳤 */
	pProcess->p_time = 0;

	if ( this->RunOut )
	{
		this->RunOut = 0;
		Kernel::Instance().GetProcessManager().WakeUpAll((unsigned long)&RunOut);
	}
}

void ProcessManager::Signal( TTy* pTTy, int signal )
{
	for ( int i = 0; i < ProcessManager::NPROC; i++ )
	{
		if ( this->process[i].p_ttyp == pTTy )
		{
			this->process[i].PSignal(signal);
		}
	}
}
